!---------------------------------------------------------------------------------------
! FLEXINVERT : bayesian 
!---------------------------------------------------------------------------------------
!  Copyright 2013, Rona Thompson
!
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!---------------------------------------------------------------------------------------
!> Bayesian inversion with inequality constraints
!! Optimal value of x:
!!
!!   x = x0 + AP^T(PAP^T)^-1(c-Px0)
!!
!! where:
!!   A  = the posterior error covariance matrix
!!   x0 = the posterior state vector without the inequality constraint
!!   P  = the constraint selector matrix
!!   c  = the constraint vector
!!
!! Ref: Thacker, Ocean Modelling, 2007 
!---------------------------------------------------------------------------------------

subroutine bayesian()

  use mod_var

  implicit none

  integer :: i,j,N,lda,info
  real, dimension(nineq) :: xerr
  real, dimension(nvar,nineq) :: tselmat
  real, dimension(nvar,nineq) :: atp,atpipatp
  real, dimension(nineq,nineq) :: patp
  real(kind=8), dimension(nineq,nineq) :: M,ipatp

  allocate(x(nvar))

  ! calculate (c-Px0)

  xerr=ineqvec-matmul(selmat,x0)

  ! calculate PAP^T

  tselmat=transpose(selmat)
  atp=matmul(cova(1:nvar,1:nvar),tselmat)
  patp=matmul(selmat,atp)

  ! calculate (PAP^T)^-1

  ! use Cholesky factorization
  N=nineq              ! order of PAP^T
  M=real(patp,kind=8)  ! symmetric matrix
  lda=nineq            ! leading dimension of PAP^T
  call dpotrf('L',N,M,lda,info)
  if(info.gt.0) write(logid,*) 'Error in calculating lower Cholesky triangle'
  call dpotri('L',N,M,lda,info)
  if(info.gt.0) write(logid,*) 'Error in inverting (HBH^T + R)'
  ! returns lower ('L') triangle of inverse of M
  ipatp=M
  do i=1,N
    do j=1,i
      ipatp(j,i)=M(i,j)
    end do
  end do

  ! constrainted state variables
  
  atpipatp=matmul(atp,real(ipatp,kind=4))
  x=x0+matmul(atpipatp,xerr)

end subroutine bayesian

